#!/usr/bin/perl

#########################################################################
#	dns/index.cgi							#
#									#
#	A script to maintain a database of DNS information and		#
#	generate DNS authority files from them.				#
#########################################################################
#
# DNS display & configuration tool 
# $Id: dns.pm,v 1.13 2001/01/17 01:57:50 jcheng Exp $
# Will DeHaan will@cobaltnet.com; Chris Caldwell, chris@nda.com

package Cobalt::UI::dns;

# use strict;
use File::Copy;
use Cobalt::Product;
use Cobalt::Netutil;
use Cobalt::Network;
use Cobalt::Locale;
use Cobalt::User;
use Locale::gettext;

#########################################################################
# Constant data
#########################################################################
# The easiest way of getting a netmask from a.b.c/nn is a table indexed by nn
my @bits_to_mask =
    (
    "0.0.0.0", "128.0.0.0", "192.0.0.0", "224.0.0.0",
    "240.0.0.0", "248.0.0.0", "252.0.0.0", "254.0.0.0",
    "255.0.0.0", "255.128.0.0", "255.192.0.0", "255.224.0.0",
    "255.240.0.0", "255.248.0.0", "255.252.0.0", "255.254.0.0",
    "255.255.0.0", "255.255.128.0", "255.255.192.0", "255.255.224.0",
    "255.255.240.0", "255.255.248.0", "255.255.252.0", "255.255.254.0",
    "255.255.255.0", "255.255.255.128", "255.255.255.192", "255.255.255.224",
    "255.255.255.240", "255.255.255.248", "255.255.255.252", "255.255.255.254",
    "255.255.255.255"
    );

# The following describes the order each field type is printed in the table.
my %typesort = (
    "soa",		"1",
    "a",		"3",
    "cname",		"4",
    "host",		"5",
    "subdomain",        "2",
    "ptr",		"7",
    "mx",		"8",
    "subnet",           "6",
    "secondary",	"10"
    );

# Order of fields in the database (columns in the file)
my @db_order = ( "type", "name", "domain", "value", "nbits" );

# Where to return to on server update
my $CallingURL = "/cgi-bin/.cobalt/turbo_ui/dispatch/services";

# Hardcoded paths for files associated with DNS
my $Named_conf = '/etc/named.conf';
my $Named_dir = '/etc/named';
my $Named_disabled = '/etc/named.disabled';
my $Named_exe = '/usr/sbin/ndc restart 2>&1 1>/dev/null';
my $Named_pid = '/var/run/named.pid';
my $Named_dnsdir = '/etc';
my $Named_cobprefix = "cobaltdns";
my $Named_cobsuffix = "conf";
my $Named_cobdns = "$Named_dnsdir/$Named_cobprefix.$Named_cobsuffix";

# Hardcoded paths associated with cobalt CGI technology
my $DNS_records = "$Named_dir/records";
my $Template_dir = '/usr/admserv/templates';

# Defaults for DNS authority files
my $Refresh = 10800;
my $Retry = 3600;
my $Expire = 604800;
my $Ttl = 86400;
my $endSOA = "; End SOA Header\n;\n; Do Not edit BIND db files directly.\n; Use the administrative web user interface /admin/ -> Control Panel -> DNS Parameters\n; Custom additions may be made by creating a file of the same name as this but with a\n; .include suffix.  Click Save Changes in the DNS web interface and the inclusion will be made.\n; \n";

#########################################################################
# Constants determined from environment					#
#########################################################################
my $iam;
#my $Debug = (-e "/etc/DEBUG") ? 1 : 0;
$^W = $Debug;

#########################################################################
# Globals								#
#########################################################################

my $user_ERROR;		# Built in do_modify, used in templatize
				# Contains errors from user's form

my @Servers;		# Built in generate_domain_contents
				# Used in add_to_tomain
				# List of name servers for all domains

my %domain_contents;	# Built in generate_domain_contents
				# Used in write_config_files
				# Text to be written to auth files for each
				# domain.

my %fields;			# Built in main from parseinfo from CGI input
				# Used all over

my %rec;			# Read in read_records, written in write_records
				# Modified and displayed all over

my($Offline);               # Flag to set wether we're editing a live
                                # DNS server, or one that hasn't been 
                                # activated.
                                # Used in restart_named()

my ($q, $locale);
#########################################################################
# Parse the CGI form and return as a hash.				#
#########################################################################
#
# This is a pretty standard CGI parsing 
sub parseinfo
    {
    my ($k,@equality,%fields);
    my $input = '';

    read(STDIN,$input,$ENV{"CONTENT_LENGTH"}) if ($ENV{"CONTENT_LENGTH"} > 0 );
    $input .= $ENV{"QUERY_STRING"}."&";

    print STDERR "Input: $input" if ($DEBUG);

    foreach $_ ( split(/&/,$input) )
	{
	/(.*)=(.*)/;
	my $left = $1;
	my $right = $2;

	$left =~ s/\%(.{2})/sprintf("%c",hex($1))/eg;
	$left =~ tr/+/ /s;

	$right =~ s/\%(.{2})/sprintf("%c",hex($1))/eg;
	$right =~ tr/+/ /s;
	$right =~ s/[#\$\*\\\^\(\)\n]/ /og;
	$right =~ s/\+/ /g;
	$right =~ s/  / /g;
	$right =~ s/\"/\"\"/g;

	if ($fields{"$left"})
	    {
	    $fields{"$left"} .= ", $right" unless ($right eq "");
	    }
	else
	    {
	    $fields{"$left"} = $right;
	    }
	}

    $fields{"remote_addr"} = $ENV{"REMOTE_ADDR"};
    $fields{"remote_host"} = $ENV{"REMOTE_HOST"};

    return(%fields);
    }


#########################################################################
# Debug writing routine.						#
#########################################################################
my $dout = 0;
sub dwrite
    {
    unlink( "/tmp/dout" ) if ($dout == 0 );
    $dout = 1;
    open( DWRITE, ">>/tmp/dout" );
    print DWRITE $_[0];
    close( DWRITE );
    }

#########################################################################
# Output the file specified in $template, each argument enclosed in	#
# in brackets with its subsequent argument.  It also automatically	#
# replaces certain strings (e.g. [IAM]) with useful values (e.g. the	#
# URL for the script.							#
#########################################################################
sub templatize
    {
    my ($template,@sublist) = @_;
    my $mask = &show_bits( $fields{"nbits"} );
    my $sm_mask = &show_lessbits( $fields{"lessbits"} );
    my $pri = &show_priority( $fields{"value"} );
    my $flav;
    my ( $i, $src, $dst );

    my $locale = Cobalt::Locale->getByHTTPHeader( $ENV{ 'HTTP_ACCEPT_LANGUAGE' } );
    my $charset = $locale->getProperty("http_Charset");
    my $charsetText;
    if ($charset) {
	$charsetText = "; charset=$charset";
    }
    print "Content-type:  text/html$charsetText\nCache-Control: no-cache\n\n";
    undef($charset, $charsetText);


    $template = "dns_$template.tmpl";

    if ($user_ERROR =~ /\w/)
	{
	$user_ERROR =~ s/^<br>//; # Crop leading <br>;
	$flav = 'error';
	}

    unless($fields{'domain'} or $fields{'cdomain'} !~ /[^\d\.\/]/)
        {
        $fields{'domain'}=$fields{'cdomain'}; # Nice to have a domain default
        }

    # Very slowly reads & displays via a template file
    if (! open(TEMPL,"$Template_dir/$template") )
	{
	print "<font color=red>Can't read template $template:  $!</font>";
	die "Can't read the template $template $!\n";
	}

    # Subnet delegation is only relevent for networks
    my $show_subnet = '<OPTION VALUE="show_subnet">[str.DNS_02_107]</OPTION>' if
      (($fields{'cdomain'} ne '/') &&($fields{'cdomain'} =~ /^[\d\.\/]+$/));
    # Subdomain delegation is only relevent for domains
    my $show_subdom = '<OPTION VALUE="show_subdomain">[str.DNS_02_106]</OPTION>'
	 if ($fields{'cdomain'} =~ /[a-zA-Z]/);

    # auto configure support
    my $auto_config = 'CHECKED' if (-e $DNS_autoconfig);

    while(<TEMPL>)
	{
	if (/\[/o)
	    {
	    $i = 0;
	    while( $sublist[$i] ne "" )
		{
		$src = $sublist[$i++];
		$dst = $sublist[$i++];
		s/\[$src\]/$dst/g;
		}
	    s/\[IAM\]/$iam/g;
	    s/\[TYPE\]/$fields{"type"}/g;
	    s/\[REC\]/$fields{"rec"}/g;
	    s/\[DOMAIN\]/$fields{"domain"}/g;
	    s/\[VALUE\]/$fields{"value"}/g;
	    s/\[HOSTNAME\]/$fields{"name"}/g;
	    s/\[NBITS\]/$fields{"nbits"}/g;
	    s/\[MASK\]/$mask/;
	    s/\[SM_MASK\]/$sm_mask/;
	    s/\[PRIORITY\]/$pri/g;

	    s/\[USER_INFO\]/\'$user_ERROR\'/g;
	    s/\[FLAV\]/$flav/g;

	    s/\[NS1\]/$fields{'ns1'}/;
	    s/\[NS2\]/$fields{'ns2'}/;
	    s/\[IP0\]/$fields{'ip0'}/;
	    s/\[IP1\]/$fields{'ip1'}/;
	    s/\[IP2\]/$fields{'ip2'}/;
	    s/\[EMAIL\]/$fields{'email'}/g;
	    s/\[REFRESH\]/$fields{'refresh'}/g;
	    s/\[RETRY\]/$fields{'retry'}/g;
	    s/\[EXPIRE\]/$fields{'expire'}/g;
	    s/\[TTL\]/$fields{'ttl'}/g;
	    s/\[AUTOCONFIG\]/$auto_config/g;
	    
	    s/\[CDOMAIN\]/$fields{'cdomain'}/g;

	    s/\[SHOW_SUBNET\]/$show_subnet/;	
	    s/\[SHOW_SUBDOM\]/$show_subdom/;	
            s/\[FONTH\]/$header_style/go;
            s/\[FONTT\]/$table_style/go;
            s/\[FONTB\]/$button_style/go;

	    s/\[str.([^\]]+)\]/Locale::gettext::dgettext("ui", "$1")/geo;
	    }
	print;
	}
    close(TEMPL);
    }


#########################################################################
# SIGHUP the name server to re-read the file (or start if not running).	#
#########################################################################
sub restart_named
    {
    if ((! $Offline) && (! -e "$Named_pid"))
	{
	system("$Named_exe");
	}
    elsif (! $Offline)
	{
	open(PID,"$Named_pid") || die("Cannot open $Named_pid:  $!");
	chomp($_ = <PID>);
	close(PID);
	system("$Named_exe");
	}
    # Otherwise, we're offline...You can't restart something that hasn't 
    # been started in the first place...wait until the named.boot file is
    # set up.
    }

#########################################################################
# This returns a host.domain allowing for a null host.			#
#########################################################################
sub append_host
    {
    return ( $_[0] eq "-" ? $_[1] : "$_[0].$_[1]" );
    }

#########################################################################
# Return the HTML in a string to replace [MASK] in the templatize	#
# routine.  The text is the "<OPTION>"s available for numbers of bits	#
# used in a network (erroneously refered to as "netmask").  It doesn't	#
# return the <select> string, just the <option> strings.		#
#########################################################################
sub show_bits
    {
    my( $actbits ) = ( ( $_[0] < 8 ) ? 24 : $_[0] );
    my( $html ) = "";

    for( $_=8; $_<=30; $_++ )
        {
	my $selflag = ( $_ == $actbits ? " selected" : "" );
	$html .= "<option value=$_ $selflag>$_\n";
	}
    return $html;
    }

#########################################################################
# Return the HTML in a string to replace [SMASK] in the templatize	#
# routine.  The text is the "<OPTION>"s available for numbers of bits	#
# used in a network (erroneously refered to as "netmask").  It doesn't	#
# return the <select> string, just the <option> strings.		#
#########################################################################
sub show_lessbits
    {
    my( $actbits ) = ( ( $_[0] < 8 ) ? 24 : $_[0] );
    my( $html ) = "";

    for( $_=$actbits; $_<=30; $_++ )
        {
	my $selflag = ( $_ == $actbits ? " selected" : "" );
	$html .= "<option value=$_ $selflag>$_\n";
	}
    return $html;
    }


#########################################################################
# Return the HTML in a string to replace [PRIORITY] in the templatize	#
# routine.  The text is the "<OPTION>"s available for MX priority.	#
# It doesn't return the <select> string, just the <option> strings.	#
#########################################################################
sub show_priority
    {
    my $pri = $_[0];
    my $html = "";

    foreach $_ ( "Very-High", "High", "Low", "Very-Low" )
	{
	my $selflag = ( ( $_ eq $pri ) ? " selected" : "" );
	if ( $_ eq "Very-High" )
	    {
	    $html .= "<option $selflag value=$_>".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_86")."\"")."\n";
	    }
        elsif ( $_ eq "High" )
	    {
	    $html .= "<option $selflag value=$_>".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_87")."\"")."\n";
	    }
        elsif ( $_ eq "Low" )
	    {
	    $html .= "<option $selflag value=$_>".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_88")."\"")."\n";
	    }
        elsif ( $_ eq "Very-Low" )
	    {
	    $html .= "<option $selflag value=$_>".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_89")."\"")."\n";
	    }
	}
    return $html;
    }

#################################################################################
# This is called with one of:							#
#	show_forward_lookup		To print form for A record		#
#	show_reverse_lookup		To print form for PTR record		#
#	show_mail_server_lookup		To print form for MX record		#
#	show_alias_lookup		To print form for CNAME record		#
#	show_secondary_forward_domain	To print form for secondary		#
#	show_secondary_reverse_domain	To print form for PTR secondary		#
#	show_soa_lookup			To print form for SOA settings		#
#	show_subnet_lookup		To print form for subnet delegation	#
#	show_subdomain_lookup		To print form for sub-domain delegation	#
# If the "rec" field is specified, we are modifying an existing record.		#
# Otherwise, this is a new record (make up a record number).			#
# Generate substitutions for templatize and then call it.			#
#################################################################################
sub show_get_data
    {
    my $templatename = $_[0];
    my ( $cflag, $modifier );
    my $rn = $fields{"rec"};
    if ($rn eq "" )
	{
	$fields{"rec"} = "new";
	$modifier = "Add";
	}
    else
	{
	my( @db_rec ) = split(/\s+/,$rec{$rn});
	for( $_=0; $db_order[$_] ne ""; $_++ )
	    {
	    $fields{ $db_order[$_] } = $db_rec[$_]
		if ($fields{$db_order[$_]} eq "" );
	    }
	$modifier = "Modify";
	}

    # This forces you to use the template that doesn't show number of
    # bits for forward lookup.  Otherwise, all would be entirely
    # symetrical.  CIAR.
    #
    if ($modifier eq "Add" )
	{
        $templatename = "add_forward_lookup" if ($_[0] eq "show_forward_lookup" );
        $templatename = "add_reverse_lookup" if ($_[0] eq "show_reverse_lookup" );
	}
    #
    if ($_[0] eq "show_forward_lookup" )
	{
	$fields{"name"} = "" if ($fields{"name"} eq "-" );
	$fields{"type"} ||= "host";
	$cflag = ( $fields{"type"} eq "host" ? "checked" : "" );
	$fields{"nbits"} = 24 if ($fields{"nbits"} <= 0 );
	$fields{"type"} = "a";
	}
    elsif ($_[0] eq "show_reverse_lookup" )
	{
	$fields{"name"} = "" if ($fields{"name"} eq "-" );
	$fields{"type"} ||= "host";
	$cflag = ( $fields{"type"} eq "host" ? "checked" : "" );
	$fields{"type"} = "ptr";
	}
    elsif ($_[0] eq "show_mail_server_lookup" )
	{
	$fields{"value"} ||= "High";
	$fields{"type"} = "mx" if ($fields{"type"} eq "" );
        if ($fields{"name"} eq '-' ) { $fields{"name"} = ''; }
	}
    elsif ($_[0] eq "show_alias_lookup" )
	{
	$fields{"type"} ||= "cname";
        if ($fields{"name"} eq '-' ) { $fields{"name"} = ''; }
        if ($fields{"value"} eq '-' ) { $fields{"value"} = ''; }
	}
    elsif ($_[0] eq "show_secondary_forward_domain" )
	{
	$fields{"type"} ||= "secondary";
	}
    elsif ($_[0] eq "show_secondary_reverse_domain" )
	{
	$fields{"type"} ||= "secondary";
	$fields{'cdomain'} = "" if ($modifier eq "Add" );
	}
    elsif ($_[0] eq "show_soa" )
	{
	$fields{'type'} ||= 'soa';
	my( @soa_elements ) = split( /:/, $fields{'value'} );
	$fields{'ns1'} ||= $soa_elements[0];
	$fields{'ns2'} ||= $soa_elements[1];
	$fields{'email'} ||= $soa_elements[2];
	$fields{'refresh'} ||= $soa_elements[3];
	$fields{'retry'} ||= $soa_elements[4];
	$fields{'expire'} ||= $soa_elements[5];
	$fields{'ttl'} ||= $soa_elements[6];
	}
    elsif ($_[0] eq "show_subdomain" )
	{
	$fields{'type'} ||= 'subdomain';
	my( @sub_elements ) = split( /:/, $fields{'value'} );
	$fields{'ns1'} ||= $sub_elements[0];
	$fields{'ns2'} ||= $sub_elements[2];
	$fields{'ip1'} ||= $sub_elements[1];
	$fields{'ip2'} ||= $sub_elements[3];
	}
    elsif ($_[0] eq "show_subnet" )
	{
	$fields{'type'} ||= 'subnet';
	my($network_baseip, $network_slash) = split(/\//, $fields{'cdomain'});
	my( @sub_elements ) = split( /:/, $fields{'value'} );
	$fields{'ns1'} ||= $sub_elements[2];
	$fields{'ns2'} ||= $sub_elements[3];
	$fields{'ip0'} ||= $sub_elements[0];
	$fields{'ip0'} ||= $network_baseip;
	$fields{'lessbits'} ||= $sub_elements[1];
	$fields{'lessbits'} ||= $network_slash + 1;
	}

    $cflag = '' if ( $modifier eq 'Add' ); # Do not automatically generate PTR by default

    &templatize( $templatename, "CALLBACK", $_[0], "FLAG", $cflag, "MODIFIER",
	( $modifier eq 'Add' ?
	    eval("\"".Locale::gettext::dgettext("ui", "DNS_02_82")."\"") :
	    eval("\"".Locale::gettext::dgettext("ui", "DNS_02_83")."\"") ) );
    }

#########################################################################
# Used by show_status for sorting records for output.  This sorts by	#
# domain first, and by type second.					#
#########################################################################
sub rec_sort
    {
    my( @atoks ) = split( /\s+/, $rec{$a} );
    my( @btoks ) = split( /\s+/, $rec{$b} );
    my $adomain = join('.',reverse(split(/\./o,$atoks[2])));
    my $bdomain = join('.',reverse(split(/\./o,$btoks[2])));
    $adomain = "zzzz$adomain" if ($atoks[0] eq "secondary" );
    $bdomain = "zzzz$bdomain" if ($btoks[0] eq "secondary" );
    my $n = ( $adomain cmp $bdomain );
    return $n if ($n );
    $n = ( $typesort{$atoks[0]} cmp $typesort{$btoks[0]} );
    return $n if ($n );
    $n = ( $atoks[1] cmp $btoks[1] );
    return $n;
    }

#########################################################################
# This returns a list of all of the records, allowing the operator	#
# to add, modify or delete them at will.				#
#########################################################################
sub show_status
    {
    my( $html, $htmldomains, $deldom ) = ('','','');
    my( $r, $type, $name, $domain, $value, $nbits, $rtn );
    my( $col1, $col2, $col3, $fcol2 );
    my $nrecs = 0;
    my $lastdomain = "";
    my @domainlist = ();
    my $domname = "";
    my %dnames = ();
    my( $selflag, $d );
    my $found_domain = 0;
    my %networks = ();
    my( $net, $mask ) = 0;

    $fields{'cdomain'} = $fields{"domain"} if ($fields{"domain"});

    if ($fields{'cdomain'} =~ /^([\d\.]+)\/(\d+)$/ )
	{
	($net, $mask) = ( $1, $2 );
	}

    foreach $r ( sort rec_sort (keys %rec) )
        {
        ( $type, $name, $domain, $value, $nbits ) = split(/\s+/,$rec{$r});

        next if (($domain eq '-') || ($type eq 'soa')); # Ignore global records

        if ($type eq 'ptr')
            {
            my( $deduced_net ) = net_get_network($value, net_convert_netmask($nbits)) .'/'. $nbits;
            $found_domain++ if ($deduced_net eq $fields{'cdomain'});
            $networks{$deduced_net} = $nbits;
	    next;
            }
	elsif ($type eq 'subnet')
	    {
            my( $deduced_net );
	    if ($domain =~ /^([\d\.]+)\/(\d+)/)
		{
                $deduced_net = net_get_network($1, net_convert_netmask($2)) .'/'. $2;
		$networks{$deduced_net} = $2;
                $found_domain++ if ($deduced_net eq $fields{'cdomain'});
		}
	    next;
	    }
	elsif (($type eq 'secondary') && ($domain !~ /[a-z]/))
	    {
	    my( $deduced_net ) = net_get_network($domain, net_convert_netmask($nbits)) .'/'. $nbits;
	    $networks{$deduced_net} = $nbits;
	    $found_domain++ if ($deduced_net eq $fields{'cdomain'});
	    next;
	    }

        if ($domain ne $lastdomain)
            {
            $domain .= '/'.$nbits if (($type eq 'secondary') && ($nbits > 0));
            $found_domain++ if ($domain eq $fields{'cdomain'});
            $dnames{$domain} = $domain;
	    push( @domainlist, $domain ) if ($domain =~ /[a-z]/);
            }

        $lastdomain = $domain;
	}

    my($numdomains) = $#domainlist + 1;
    $numdomains = 10 if ($numdomains > 10 );
    @domainlist = sort(@domainlist);

    unless( $found_domain )
	{
	$fields{'cdomain'} = $fields{'domain'} = (keys %networks)[0];
    	$fields{'cdomain'} = $fields{'domain'} = $domainlist[0] if ($domainlist[0]);
	}
    
    if ($fields{'cdomain'} =~ /^([\d\.]+)\/(\d+)$/ )
	{
	($net, $mask) = ( $1, $2 );
	}

    if (($#domainlist < 0) && length((keys %networks) < 1))
	{
	$html .= "<table BORDER=1 width=600 CELLSPACING=0 CELLPADDING=0>";
	$html .= "<tr><td bgcolor=\"#000099\" colspan=5>";
	$html .= "<FONT STYLE=$header_style>";
	$html .= "<b>".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_85")."\"")."</b></font></td></tr>";
	$html .= "<tr><td align=left colspan=5>&nbsp;<i>".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_58")."\"")."</i></td></tr>\n";
	#$htmldomains .= "<select valign=center>";
	#$htmldomains .= "<option>".eval("\"".Locale::gettext::dgettext("ui", "DNS_04_11")."\"")."</option>\n";
	#$htmldomains .= "</select>";
	}
    else
	{
	my( $first_pass ) = 1;
	foreach $r ( sort rec_sort (keys %rec) )
	    {
	    # This toggles between white and gray for each record we print out.
	    # Note that "host" records take two lines (one to show forward
	    # map, one to show reverse map).
	    # my $bgcolor = ( $nrecs % 2 ? "#ffffff" : "#e8e8e8" );
	    my $bgcolor = "white";

	    my @typelist;

	    next unless($rec{$r});	# Ignore empty records

	    ($type, $name, $domain, $value, $nbits) = split(/\s+/,$rec{$r});

	    if ($net && ($type eq 'ptr'))
		{
		next unless netutil_test_network_in_network( $net, $mask, $value, $nbits );
		next unless ($nbits eq $mask);
		}
	    elsif (($type eq 'secondary') && ($domain !~ /[a-z]/))
		{
	        next unless netutil_test_network_in_network( $net, $mask, $domain, $nbits );
                next unless ($nbits eq $mask);
		}
            elsif (($type eq 'soa') && ($fields{'cdomain'} =~ /^[\d\.]+\/\d+$/) &&
                   ($domain =~ /^([\d\.]+)\/(\d+)$/))
		{ # soa for network authority
                my($soa_net, $soa_sn) = ($1, $2);
	        next unless netutil_test_network_in_network( $net, $mask, $soa_net, $soa_sn );
		next if ($soa_sn ne $mask); # avoid network-subnet authority conflicts
		}
	    elsif ($type eq 'soa')
		{ # soa for forward domain
		next if ($domain ne $fields{'cdomain'});
		}
	    elsif ($domain ne $fields{'cdomain'})
		{
	        next;
		}
	    if ($first_pass )
		{
		$html .= "</table><p>" if ($lastdomain );
		$html .= "<table BORDER=1 width=600 CELLSPACING=0 CELLPADDING=0>";
		$html .= "<tr><td colspan=4 bgcolor=\"#000099\" align=center>";
		$html .= "<table><tr>";
		$html .= "<th valign=middle bgcolor=\"#000099\" nowrap>";
		$html .= "<FONT STYLE=$header_style>";
		$html .= "<b>".eval("\"".Locale::gettext::dgettext("ui", "DNS_03_85")."\"")." ";
		if ($fields{'cdomain'} !~ /\// )
		    {
		    $html .= $dnames{$domain};
		    }
		else
		    {
		    $html .= $fields{'cdomain'};
		    }
		    
		$htmldomains .= "<select valign=center size=1 name=cdomain ";
		$htmldomains .= "onchange=\"document.sd.submit();\">\n";
		$htmldomains .= "<option VALUE=\"".$fields{'cdomain'}."\">".eval("\"".Locale::gettext::dgettext("ui", "DNS_04_11")."\"")."</option>\n";
		
		foreach $d (@domainlist)
		    {
		    # $selflag=( $d eq $domain ? "selected" : "" );
		    $htmldomains .= "<option value=$d $selflag>$dnames{$d}</option>\n";
		    }
		foreach $d ( sort( keys( %networks ) ) )
		    {
		    $htmldomains .= "<option value=\"$d\">$d</option>\n";
		    }
		$htmldomains .= "</select>";
		$html .= "</b></font></td>";
		$html .= "</tr></table>";

		# UI feature to delete all domain record types in a single pass
		# This is useful for forward domains but is less desirable
		# for reverse networks
                my $deldomhelp=Locale::gettext::dgettext("ui", "DNS_04_12");
                $deldom=<<END;
<TABLE BORDER="0" CELLPADDING="0" CELLSPACING="0"><TR>
  <TD><A HREF="javascript: delDomain('$dnames{$domain}');" onMouseOver="status='$deldomhelp'; return true;"><IMG SRC="/libImage/buttonLeftDot.gif" BORDER=0></A></TD>

  <TD NOWRAP STYLE="[FONTB]" ><A HREF="javascript: delDomain('$dnames{$domain}');" onMouseOver="status='$deldomhelp'; return true;"><FONT STYLE="[FONTB]">[str.delDomainbut]</FONT></A></TD>

  <TD><A HREF="javascript: delDomain('$dnames{$domain}');" onMouseOver="status='$deldomhelp'; return true;"><IMG SRC="/libImage/buttonRight.gif" BORDER=0></A></TD>

</TR></TABLE>
<BR>
END

		$deldom = '' if ($fields{'cdomain'} =~ /^[\d\.\/]+$/ );

		$first_pass = 0;
		}

		if ($type eq "a" )		# Set fields for table by type
		    {
		    $col1 = &append_host( $name, $domain );
		    $col2 = "<img src=/.cobalt/images/RightArrow alt=\"".eval("\"".Locale::gettext::dgettext("ui", "DNS_04_13")."\"")."\">";
		    $col3 = "$value";
		    $rtn = "show_forward_lookup";
		    }
		elsif ($type eq "ptr" )
		    {
		    $col1 = "$value/$nbits";
		    $col2 = "<img src=/.cobalt/images/RightArrow alt=\"".eval("\"".Locale::gettext::dgettext("ui", "DNS_04_13")."\"")."\">";
		    $col3 = &append_host( $name, $domain );
		    $rtn = "show_reverse_lookup";
		    }
		elsif ($type eq "mx" )
		    {
		    $col1 = &append_host( $name, $domain );

		      if ($value eq "Very-Low" )
			    { $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_03_101")."\""); }
		      elsif ($value eq "Low" )
			    { $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_03_100")."\""); }
		      elsif ($value eq "High" )
			    { $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_03_99")."\""); }
		      elsif ($value eq "Very-High" )
			    { $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_03_98")."\""); }

		    $col3 = "$nbits";

		    $rtn = "show_mail_server_lookup";
		    }
		elsif ($type eq "cname" )
		    {
		    $col1 = &append_host( $name, $domain );
		    $col2 = "<img src=/.cobalt/images/RightArrow alt=\"".eval("\"".Locale::gettext::dgettext("ui", "DNS_04_13")."\"")."\">";
		    $col3 = &append_host( $value, $nbits );
		    $rtn = "show_alias_lookup";
		    }
		elsif ($type eq "secondary" )
		    {
		    if ($nbits >= 8 )
			{
			$col1 = "$domain/$nbits";
			$rtn = "show_secondary_reverse_domain";
			}
		    else
			{
			$col1 = "$domain";
			$rtn = "show_secondary_forward_domain";
			}
		    $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_02_97")."\"");
		    $col3 = "$value";
		    }
		elsif ($type eq "soa" )
		    {
		    $col1 = $domain;
		    $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_02_108")."\"");
		    $col3 = '';
		    $rtn = "show_soa";
		    }
		elsif ($type eq "subdomain" )
		    {
		    my( $ns1, $ip1, $ns2, $ip2 ) = split( /:/, $value );
		    $col1 = &append_host( $name, $domain );
		    $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_02_134")."\"");
		    $col3 = "$ns1";
		    if ($ns2 )
			{
		        $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_02_143")."\"");
		        $col3 = "$ns1<BR><FONT COLOR=#FFFFFF>&nbsp;</FONT>$ns2"; 
			}
		    $rtn = "show_subdomain";
		    }
		elsif ($type eq "subnet" )
		    {
		    my( $network_baseip, $network_subnet, $ns1, $ns2 ) = split( /:/, $value );
		    $col1 = $network_baseip.'/'.$network_subnet;
		    $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_04_32")."\"");
		    $col3 = "$ns1";
		    if ($ns2 )
			{
		        $col2 = eval("\"".Locale::gettext::dgettext("ui", "DNS_04_33")."\"");
		        $col3 = "$ns1<BR><FONT COLOR=#FFFFFF>&nbsp;</FONT>$ns2"; 
			}
		    $rtn = "show_subnet";
		    }
		else
		    {
		    $rtn = "unknown";
		    $col1 = "";
		    $col2 = "";
		    $col3 = "";
		    }

                # Special case for SOA records, they don't need three columns
                if ( $type eq "soa" )
		    {
	  	    # Build up SOA string of a table row
		    my $DNS_02_108=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_108")."\"");
		    my $BCC_051=eval("\"".Locale::gettext::dgettext("ui", "BCC_051")."\"");
		    $html .= <<EOF;
<tr><td colspan=3 bgcolor=$bgcolor width=20% align=center>$DNS_02_108</td>
<td bgcolor=$bgcolor align=center valign=middle width=10%>&nbsp;
<a href="$iam?cdomain=$fields{cdomain}&rec=$r&func=$rtn" onMouseOver="window.status='$BCC_051 $col1 $DNS_02_108'; return true;"><img src="/libImage/modify" border=0></a>&nbsp;
<IMG SRC="/.cobalt/images/white_color" BORDER=0>
</td>
</tr>
EOF
                    }
                else 
		    {
	  	    # Build up HTML string of a table row
		    
		    my $alt_text = $col3;
		    $alt_text =~ s/\<.+\>/ /g; # Clean up alt text

		    $html.= "<tr><td bgcolor=$bgcolor width=20%>&nbsp;$col1&nbsp;</td>";
		    $html.= "<td bgcolor=$bgcolor align=center width=50%>&nbsp;$col2&nbsp;</td>";
		    $html.= "<td bgcolor=$bgcolor width=20%>&nbsp;$col3&nbsp;</td>";
		    $html .= "<td bgcolor=$bgcolor align=center valign=middle width=10%>";
		    if ($col2 =~ />/ )
		        { $fcol2 = "-"; }
		    else
		        { $fcol2 = $col2; }
		    $html .= "<a href=\"$iam?cdomain=$fields{cdomain}&rec=$r&func=$rtn\" onMouseOver=\"window.status='".eval("\"".Locale::gettext::dgettext("ui", "BCC_051")."\"")." \\'$col1 $fcol2 $alt_text\\' ".eval("\"".Locale::gettext::dgettext("ui", "BCC_052")."\"")."'; return true;\">";
		    $html .= "<img alt=\"".eval("\"".Locale::gettext::dgettext("ui", "BCC_051")."\"")." '$col1 $fcol2 $alt_text' ".eval("\"".Locale::gettext::dgettext("ui", "BCC_052")."\"")."\" src=\"/libImage/modify\" border=0\">";
		    $html .= "</a>&nbsp;";
		    $html .= "<a href=\"javascript: delEntry('$type',$r);\" onMouseOver=\"window.status='".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_65")."\"")."'; return true;\">";
		    $html .= "<img alt=\"".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_65")."\"")."\" src=\"/libImage/remove\" border=0>";
		    $html .= "</a></td>";
		    $html .= "</tr>\n";
                    }
		}
	    $nrecs++;
	}

    $html .= "</table>"; # if ($lastdomain ne "" );

    &templatize( "root", "REPLACE", $html, "LISTDOMAIN", $htmldomains,
		 "DELDOM", $deldom );
    }

#########################################################################
# Add a line to to the string being built up for each domain.  If the	#
# domain has nothing in it, add a default SOA record.			#
#########################################################################
sub add_to_domain
    {
    my( $name, $domain, $type, $value ) = @_;
    my( $s );
    
    my($local_domain) = $domain;    
    $local_domain = (&zone_to_domain_list( $domain ))[0] if ($domain =~ /\//);
    $_ = $domain_contents{$local_domain};

    # Forward lookups use customizable SOA
    if ($type eq 'soa' )
	{
	my( $ns1, $ns2, $email, $refresh, $retry, $expire, $ttl ) = split( /:/, $value );
	$ns1 ||= $Servers[0]; # Assign defaults
        $email ||= 'admin@'.$Servers[0];
        $email =~ s/\@/\./; # BIND8 treats the first . as the @ in the rp record
	$refresh ||= $Refresh;
	$retry ||= $Retry;
        $expire ||= $Expire;
        $ttl ||= $Ttl;

	my( $soa_record ) = <<EOF;
\$TTL $ttl
$local_domain. IN SOA $ns1. $email. (
	::SERIAL::
        $refresh
        $retry
        $expire
	$ttl
        )
EOF
	# It would be nice to auto-gen glue records here, but we need IP addrs for the NSs
	$soa_record .= "$local_domain.	IN	NS	$ns1.\n";
	$soa_record .= "$local_domain.	IN	NS	$ns2.\n" if ( $ns2 );
	$soa_record .= $endSOA;
	
        $_ = $soa_record . $_;    # Prepend updated SOA
	}
    elsif ( $type eq 'subdomain' )
	{
	my( $ns1, $ip1, $ns2, $ip2 ) = split( /:/, $value );
	$_ .= "$name.$domain.	IN	NS	$ns1.\n";
	$_ .= "$name.$domain.	IN	NS	$ns2.\n" if ( $ns2 );
	$_ .= "$ns1.	IN	A	$ip1.\n" if ( $ip1 ); # Glue records
	$_ .= "$ns2.	IN	A	$ip2.\n" if ( $ip1 && $ns2);
	}
    elsif ( $type eq 'subnet' )
	{
	# subnet - 216.65.212.0/24 216.65.212.12:25:ns1.subnetserver.net:ns2.subnetserver.net -
	my($net_baseip, $net_slash, $ns1, $ns2) = split( /:/, $value );
	my(@ip_frag) = split(/\./, $net_baseip);
	if ($net_slash == 16) # Octet Class B
	    {
	    $_ .= "$ip_frag[1]\t$Ttl\tIN\tNS\t$ns1.\n";
	    $_ .= "$ip_frag[1]\t$Ttl\tIN\tNS\t$ns2.\n" if ($ns2);
	    }
	elsif ($net_slash == 24) # Octet Class C
	    {
	    $_ .= "$ip_frag[2]\t$Ttl\tIN\tNS\t$ns1.\n";
	    $_ .= "$ip_frag[2]\t$Ttl\tIN\tNS\t$ns2.\n" if ($ns2);
	    }
	else			 # Non-octect bounded
	    {
	    my $node_base = netutil_get_network($net_baseip, $bits_to_mask[$net_slash]);
	    unless($net_baseip eq $node_base)
		{
		$net_baseip = $node_base;
		}
	    
	    my ($node, $node_low, $diff_mask, $for_node, $for_net);

	    if ($net_slash > 23)
		{
		$node_base =~ s/\.\d+$/\.REP/;
		$node_low = (split(/\./,$net_baseip))[3];
		$diff_mask = 32;
		}
	    elsif ($net_slash > 15)
		{
		$node_base =~ s/\.\d+\.\d+$/\.REP/;
		$node_low = (split(/\./,$net_baseip))[2];
		$diff_mask = 24;
		}
	    elsif ($net_slash > 7)
		{
		$node_base =~ s/^(\d+\.)\d+\..*/$1REP/;
		$node_low = (split(/\./,$net_baseip))[1];
		$diff_mask = 16;
		}
	    else
		{
		$node_base = 'REP';
		$node_low = (split(/\./,$net_baseip))[0];
		$diff_mask = 8;
		}

 	    $for_node = $node_base;
	    my $node_hi = $node_low+2**($diff_mask-$net_slash) - 1;
	    my $node_range = $node_low.'/'.$net_slash;
	    $for_node =~ s/REP/$node_range/;
	    $node_range = join('.', reverse( split(/\./, $for_node) ) ).'.in.addr.arpa.';
		
	    $_ .= "; RFC 2317 Compliant Subnet Delegation  <<$node_low-$node_hi>> /$net_slash\n";
	    $_ .= "$node_low/$net_slash\tIN\tNS\t$ns1.\n";
	    $_ .= "$node_low/$net_slash\tIN\tNS\t$ns2.\n" if ($ns2);
	    for ($node = $node_low + 1; 
	         $node < $node_low+2**($diff_mask-$net_slash); $node++)
		{
		# Simple octet-style, works if client is non-rfc2317
		$for_node = $node_base;
		$for_node =~ s/REP/$node/;

		$for_net = join('.', reverse( split(/\./, $for_node) ) ).'.in.addr.arpa.';
		$_ .= "$node\tIN\tCNAME\t$node\.$node_range\n";
		}
	    $_ .= ";\n"; # db Readability
	    }
	}
    else
    	{
    	$_ .= "$name\tin\t$type\t$value\n";
	}
    
    $domain_contents{$local_domain} = $_;
    }

#########################################################################
# Create %domain_contents, a hash of domains with the strings to be	#
# put in each domain's authority file.  We do create pseudo entries for	#
# secondary domains, but write_config_files will figure out what to do.	#
#########################################################################
sub generate_domain_contents
    {
    my( $k, $type, $name, $domain, $value, $nbits );

    undef %domain_contents;
 
    open(NAME,"/proc/sys/kernel/hostname") || die "No hostname: $!";
    chomp($Servers[0] = <NAME>);
    close NAME;

    &add_to_domain("1","0.0.127.in-addr.arpa","ptr","localhost.");
    &add_to_domain("1","0.0.127.in-addr.arpa","soa","localhost::admin.localhost:$Refresh:$Retry:$Expire:$Ttl");

    # Go through each record, figure out what domain it is in and
    # add to the appropriate record.
    foreach $_ ( values %rec )
        {
	next if ($_ eq "" );
	( $type, $name, $domain, $value, $nbits ) = split(/\s+/);

	if ($type eq "secondary" )
	    {
	    $domain_contents{$domain} = "secondary $name $value $nbits";
	    }
	elsif ($type eq "a" || $type eq "host" )
	    {
	    $_ = &append_host( $name, $domain ) . ".";
	    if ($type eq "host" )
		{
		&add_to_domain( &ip_to_revname($value,$nbits),
		    &ip_to_domain($value,$nbits), "ptr", $_ );
		}
	    &add_to_domain( $_, $domain, "a", $value );
	    }
	elsif ($type eq "ptr" )
	    {
	    &add_to_domain( &ip_to_revname($value,$nbits),
		&ip_to_domain($value,$nbits), "ptr", &append_host($name,$domain.".") );
	    }
	elsif ($type eq "mx" )
	    {
	    if ($value eq "Very-Low" )
	        { $value = "50"; }
	    elsif ($value eq "Low" )
	        { $value = "40"; }
	    elsif ($value eq "High" )
	        { $value = "30"; }
	    elsif ($value eq "Very-High" )
	        { $value = "20"; }
	    else
	        { $value = "10"; }
	    if ($name eq "" || $name eq "-" )
		{
		&add_to_domain( "$domain.", $domain, "mx", "$value $nbits." );
		}
	    else
		{
		&add_to_domain( $name, $domain, "mx", "$value $nbits." );
		}
	    }
	elsif ($type eq "cname" )
	    {
	    if ($name eq "" || $name eq "-" )
		{
		&add_to_domain( "$domain.", $domain, $type, &append_host($value,$nbits."." ) );
		}
	    else
		{
		&add_to_domain( $name, $domain, $type, &append_host($value,$nbits."." ) );
		}
	    }
	elsif ($type eq "soa" )
	    {
	    if ($domain =~ /^([\d\.]+)\/(\d+)$/ )
	   	{
		my( $network, $mask ) = ( $1, $2 );
	    	&add_to_domain( '-', &ip_to_domain($network,$mask), 'soa', $value );
		}
	    else
	   	{
	    	&add_to_domain( '-', $domain, $type, $value );
		}
	    }
	elsif ($type eq "subnet" )
	    {
	    my( $net, $net_slash ) = split(/\//, $domain );
	    &add_to_domain( $name, &ip_to_domain($net, $net_slash), $type, $value, $_ );
	    }
	else
	    {
	    &add_to_domain($name, $domain, $type, $value);
	    }
	}

    }

#########################################################################
# Return a list of domains in various different formats based on the	#
# /etc/cobaltdns.conf file and the actual domain.			#
#########################################################################
sub zone_to_domain_list
    {
    my $domain = $_[0];
    my @best_ret = ();
    my @sec_ret = ();
    $_ = $domain;
    /\/(\d+)/;
    my $nbits = $1;
    s/\/\d+//;
    s/\.in-addr\.arpa//;
    s/^/0.0.0.0./;
    my( @ip ) = reverse( split(/\./) );

    open( NCD, $Named_cobdns ) || return $domain;
    while( $_ = <NCD> )
        {
	chomp;
	my( $lookfor, $returnpat ) = split(/\s+/);
	if ($lookfor eq $domain )
	    { push( @best_ret, $returnpat ); }
	elsif(
	    ( $lookfor eq "zone-format-24:" && $nbits > 24 )
	||  ( $lookfor eq "zone-format-16:" && $nbits > 16 && $nbits < 24 )
	||  ( $lookfor eq "zone-format-8:" && $nbits > 8 && $nbits < 16 ) )
	    {
	    $returnpat =~ s/%1/$ip[0]/;
	    $returnpat =~ s/%2/$ip[1]/;
	    $returnpat =~ s/%3/$ip[2]/;
	    $returnpat =~ s/%4/$ip[3]/;
	    $returnpat =~ s/%n/$nbits/;
	    push( @sec_ret, $returnpat );
	    }
	}
    close( NCD );
    return @best_ret if (@best_ret );
    return @sec_ret if (@sec_ret );
    return ( $domain );
    }

#########################################################################
# Remove files in named that we won't need.				#
#########################################################################
sub clean_up_dir
    {
    my $file;
    chdir( $Named_dir ) || die("Cannot chdir($Named_dir):  $!");
    opendir( NDD, "." ) || die("Cannot open $Named_dir:  $!");
    while( $file = readdir(NDD) )
	{
	next if ($file =~ /^sec\./ );
	next if ($file eq "db.cache" );
	next if ($file eq "records" );
	next if ($file =~ /\.include$/ );
	next if ($file =~ /^pri\.(.*)/ and $file eq domain_to_file_name($1));
	unlink( $file );
	}
    closedir( NDD );
    }

#########################################################################
# Create authority files from %domain_contents hash.  Also create	#
# named.conf.  Tries to be smart:  If authority file hasn't changed,	#
# do not increment the serial number.					#
#########################################################################
sub write_config_files
    {
    my($sec,$min,$hour,$mday,$month,$year) = localtime(time);
    my( $changed_domains );
    $month++;
    $year += ( ($year < 80) ? 2000 : 1900 );
    my($gennedsn) = sprintf("%04d%02d%02d%02d",$year,$month,$mday,$hour);
    
    my( $domain, $fname, $contents, $prevcontents, $serial );

    chdir( $Named_dir ) || die("Cannot chdir($Named_dir):  $!");

    my( $forward0, $forward1, $axs, $r, @rec_data );
    foreach $r ( keys %rec )
	{
	@rec_data = split(/\s+/, $rec{$r});
	next unless ( $rec_data[2] eq '-' ); # Global only
	$forward0 = $rec_data[3] if ( $rec_data[0] eq 'forwarder0' );
	$forward1 = $rec_data[3] if ( $rec_data[0] eq 'forwarder1' );
	$axs = $rec_data[3] if ( $rec_data[0] eq 'axs' );
	}
    $axs =~ s/\:\:/; /g;
    $axs =~ s/\s*\;\s*$//g; # flush null xfer hosts

    # avoid race
    my $Named_lock = "$Lockdir/named.conf";
    sleep 2 if (-e $Named_lock);
    sleep 2 if (-e $Named_lock);
    sleep 2 if (-e $Named_lock);
    sleep 2 if (-e $Named_lock);

    open( NAMEDCONF, ">$Named_lock" ) || die("Cannot write $Named_lock:  $!");
    print NAMEDCONF "options {\n\tdirectory \"$Named_dir\";\n";
    if ($forward0 && $forward1 )
	{
	print NAMEDCONF "\tforwarders { $forward0; $forward1; };\n";
	}
    elsif ($forward0 || $forward1 )
	{
	print NAMEDCONF "\tforwarders { $forward0$forward1; };\n";
	}
    if ($axs)
	{
	print NAMEDCONF "\tallow-transfer { $axs; };\n";
	}
    print NAMEDCONF "};\n";
    print NAMEDCONF "zone \".\" { type hint; file \"db.cache\"; };\n";

    # Loop through each domain, create the file and dump its contents.
    $changed_domains = 0;
    foreach $domain ( keys %domain_contents )
        {
	next unless( $domain );
	next if ($domain eq '-' ); # Ignore global settings
        $fname = &domain_to_file_name($domain);

	if ($domain_contents{$domain} =~ /secondary ([^ ]+) ([^ ]+) ([^ ]+)/ )
	    {
	    my ( $name, $value, $nbits ) = ( $1, $2, $3 );

	    if ($nbits <= 0 )
		{
		print NAMEDCONF
"zone \"$domain\" { type slave; file \"$fname\"; masters { $value; }; };\n";
		}
	    else
		{
		my @zlist = &zone_to_domain_list(&ip_to_domain($domain,$nbits));
		my $dind;
		foreach $dind ( @zlist )
		    {
		    print NAMEDCONF
"zone \"$dind\" { type slave; file \"$fname\"; masters { $value; }; };\n";
		    }
		}
	    next;
	    }
	
	if ($domain !~ /\// )
	    {
	    print NAMEDCONF
		"zone \"$domain\" { type master; file \"$fname\"; };\n";
	    }
	else
	    {
	    foreach $_ ( &zone_to_domain_list( $domain ) )
	        {
		print NAMEDCONF
		    "zone \"$_\" { type master; file \"$fname\"; };\n";
		}
	    }

	$contents = $domain_contents{$domain};
	
	if (! open( INF, $fname ) )
	    {
	    $serial = $gennedsn;
	    }
	else
	    {
	    $prevcontents = "";
	    while( $_ = <INF> )
	        {
		$prevcontents .= $_;
		}
	    close( INF );
	    $_ = $prevcontents;
	    if (/\(\s*(\d+)\b/ )
		{
	        $serial = $1;
		}
	    else
	        {
		$serial = 0;
		}
	    $prevcontents =~ s/\(\s*(\d+)\b/( ::SERIAL:: /;
	    next if ($contents eq $prevcontents );
	    $serial++;
	    $serial = $gennedsn if ($serial <= $gennedsn );
	    }
	$contents =~ s/::SERIAL::/$serial/;
	open( OUT, "> $fname" ) || die("Cannot write to $fname:  $!");
	print OUT "$contents";
	# Tolerate customizations, RP, TXT's, etc.
	print OUT "\$INCLUDE $fname.include\n" if (-r "$fname.include" );
	close( OUT );
	$changed_domains++;
	}
    close( NAMEDCONF );
    rename($Named_lock, $Named_conf);

    return $changed_domains;
    }

#########################################################################
# Map domain name to file name						#
#########################################################################
sub domain_to_file_name
    {
    $_ = $_[0];
    $_ =~ tr/\//-/;
    return undef unless (exists $domain_contents{$_[0]});
    return "sec.$_" if ($domain_contents{$_[0]} =~ /secondary/ );
    return "pri.$_";
    }

#########################################################################
# Convert IP address to form appropriate to put in revhost auth file.	#
# For instance, 1.2.3.4 in an 8 bit network would be converted to 4.3.2	#
#########################################################################
sub ip_to_revname
    {
    my( $ip, $nbits ) = @_;
    my( @ipa ) = split( /\./, $ip );
    my( @maska ) = split( /\./, $bits_to_mask[$nbits] );
    my( $res ) = "";
    my( $i );
    for( $i=3; $i>=0; $i-- )
	{
	if ($maska[$i] != 255 )
	    {
	    $res .= "." if ($res ne "" );
	    $res .= $ipa[$i];
	    }
	}
    return $res;
    }

#########################################################################
# Convert IP address and number of bits to a domain name.  For instance	#
# 192.168.1.70 in 26 bits would become 64/26.1.168.192.in-addr.arpa.	#
# From RFC 2317.							#
#########################################################################
sub ip_to_domain
    {
    my( $ip, $nbits ) = @_;
    my( @ipa ) = split( /\./, $ip );
    my( @maska ) = split( /\./, $bits_to_mask[$nbits] );
    my( $res ) = "";
    my( $i );
    for( $i=3; $i>=0; $i-- )
	{
	if ($maska[$i] != 0 )
	    {
	    $res .= "." if ($res ne "" );
	    $res .= ( ($maska[$i]+0) & ($ipa[$i]+0) );
	    $res .= "/$nbits" if ($maska[$i] ne 255 );
	    }
	}
    return "$res.in-addr.arpa";
    }

#########################################################################
# Loop through each record looking for records that match argument	#
# (a string of fields separated with spaces with * being a wildcard).	#
#########################################################################
sub match
    {
    my @matcharray = split(/\s+/,$_[0]);
    my $ri;
    foreach $ri ( keys %rec )
        {
	next if ($ri eq $fields{rec} );
	my @check = split( /\s+/, $rec{$ri} );
	for( $_=0; $check[$_]; $_++ )
	    {
	    next if ($matcharray[$_] eq "*" );
	    last if ($matcharray[$_] ne $check[$_] );
	    }
	return 1 if ($matcharray[$_] eq "" );
	}
    return 0;
    }

#########################################################################
# Do error checking on form input.  If no errors, replace/add the rec.	#
#########################################################################
sub do_modify
    {
    my $type = ( ( $fields{"ishost"} ne "" ) ? "host" : $fields{"type"} );
    my ( $f, $t );
    my @typelist;

    $user_ERROR = "";

    $fields{"type"} =~ tr/A-Z/a-z/;	# Eliminate case fumbles
    $fields{"name"} =~ tr/A-Z/a-z/;
    $fields{"domain"} =~ tr/A-Z/a-z/;

    if ($type eq "a" || $type eq "host" || $type eq "ptr" )
        {
	if ($fields{"name"} eq "" || $fields{"name"} eq "-" )
	    {
	    $fields{"name"} = "-";
	    }
	else
	    {
	    $user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_49")."\"")
		if ($fields{"name"} !~ /^[A-Za-z0-9-]+(\.[A-Za-z0-9-]+)*$/ );
	    }

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_11")."\"")
	    if (! Cobalt::Netutil::netutil_test_domainname($fields{"domain"}) );

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_12")."\"")
	    if (! Cobalt::Netutil::netutil_test_ipaddr($fields{"value"}) );
	$user_ERROR.="Number bits ($fields{nbits}) is not between 8 and 30.  "
	    if ($fields{"nbits"} < 8 || $fields{"nbits"} > 30 );
	$user_ERROR.="$fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_93")."\"").""
	    if (&match("secondary * $fields{domain} * *") && $type ne "ptr" && $fields{domain} );
	if ($type eq "host" )
	    {
	    $user_ERROR.="$fields{name}.$fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_90")."\"")." $fields{value} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_91")."\"").""
	        if (( &match("a $fields{name} $fields{domain} $fields{value} *")
	             ||  &match("ptr $fields{name} $fields{domain} $fields{value} *")
	             ||  &match("host $fields{name} $fields{domain} $fields{value} *")
		    )	
                && $fields{value}
	        ) ;
	    }
	elsif ($type eq "a" )
	    {
	    $user_ERROR.="$fields{name}.$fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_90")."\"")." $fields{value} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_91")."\"").""
	        if ((&match("a $fields{name} $fields{domain} $fields{value} *")
	             ||  &match("host $fields{name} $fields{domain} $fields{value} *")
		     )	
 		 && $fields{value} ) ;
	    }
	elsif ($type eq "ptr" )
	    {
	    $user_ERROR.="$fields{name}.$fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_90")."\"")." $fields{value} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_91")."\"").""
	        if (&match("ptr $fields{name} $fields{domain} $fields{value} *")
	        ||  &match("host $fields{name} $fields{domain} $fields{value} *") 		 && $fields{value}) ;
	    }
	$user_ERROR.="$fields{value}/$fields{nbits} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_93")."\"").""
	    if (&match("secondary * * $fields{value} $fields{nbits}") &&
	        $type ne "a" && $fields{value} ne "" );
	$fields{"type"} = $type;
	}
    elsif ($type eq "mx" )
	{
	if ($fields{"name"} ne "" )
	    {
	    $user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_50")."\"")
		if ($fields{"name"} !~ /^[A-Za-z0-9-]+(\.[A-Za-z0-9-]+)*$/ );
	    }

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_11")."\"")
	    if (! Cobalt::Netutil::netutil_test_domainname($fields{"domain"}) );

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_48")."\"")
	    if ($fields{"value"} ne "Very-Low"	&&
		$fields{"value"} ne "Low"	&&
		$fields{"value"} ne "High"	&&
		$fields{"value"} ne "Very-High"	);

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_50")."\"")
	    if (! Cobalt::Netutil::netutil_test_domainname($fields{"nbits"}) );

	$user_ERROR.="$fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_93")."\"").""
	    if (&match("secondary * $fields{domain} * *") && $fields{domain} );

	$user_ERROR.="$fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_94")."\"").""
            if (&match("mx $fields{name} $fields{domain} * $fields{nbits}") && $fields{domain} );

	}
    elsif ($type eq "cname" )
	{
	if ($fields{"name"} ne "" )
	    {
	    $user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_50")."\"")
		if ($fields{"name"} !~ /^[A-Za-z0-9-]+(\.[A-Za-z0-9-]+)*$/ );
	    }

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_11")."\"")
	    if (! Cobalt::Netutil::netutil_test_domainname($fields{"domain"}) );

	if ($fields{"value"} ne "" )
	    {
	    $user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_50")."\"")
		if ($fields{"value"} !~ /^[A-Za-z0-9-]+(\.[A-Za-z0-9-]+)*$/ );
	    }

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_11")."\"")
	    if (! Cobalt::Netutil::netutil_test_domainname($fields{"nbits"}) );

	$user_ERROR.="$fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_93")."\"").""
	    if (&match("secondary * $fields{domain} * *") && $fields{domain} );

        $user_ERROR.="$fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_94")."\"").""
            if (&match("mx $fields{name} $fields{domain} * *") && $fields{domain} );

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_04_15")."\"")
	    if (&match("cname $fields{name} $fields{domain} * *") && $fields{domain} );
	}
    elsif ($type eq "secondary" && $fields{"nbits"} <= 0 )
        {
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_18")."\"")
	    if ($fields{"name"} ne "" );

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_11")."\"")
	    if (! Cobalt::Netutil::netutil_test_domainname($fields{"domain"}) );

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_12")."\"")
	    if (! Cobalt::Netutil::netutil_test_ipaddr($fields{"value"}) );

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_55")."\"")
	    if ($fields{"name"} ne "" );
	$user_ERROR.="".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_95")."\"")." $fields{domain} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_96")."\"").""
	    if (&match("* * $fields{domain} * *") && $fields{domain} );
	}
    elsif ($type eq "secondary" && $fields{"nbits"} > 0 )
        {

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_18")."\"")
	    if ($fields{"name"} ne "" );

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_51")."\"")
	    if (! &isValidNetworkName($fields{"domain"}) );

	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_12")."\"")
	    if (! Cobalt::Netutil::netutil_test_ipaddr($fields{"value"}) );
	$user_ERROR.="".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_19")."\"")." "
	    if ($fields{"nbits"} < 8 || $fields{"nbits"} > 30 );
	$user_ERROR.="".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_95")."\"")." $fields{domain}/$fields{nbits} ".eval("\"".Locale::gettext::dgettext("ui", "DNS_02_96")."\"").""
	    if (&match("* * $fields{domain} * $fields{nbits}") && $fields{domain});
	}
    elsif ($type eq 'soa' )
	{
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_129")."\"")
	    unless( Cobalt::Netutil::netutil_test_domainname($fields{"ns1"}) );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_129")."\"")
	    if ($fields{"ns2"} && !( Cobalt::Netutil::netutil_test_domainname($fields{"ns2"}) ) );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_125")."\"")
	    unless( Cobalt::Netutil::netutil_test_emailaddr($fields{"email"}) );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_130")."\"")
	    unless( $fields{"refresh"} =~ /^\d+$/ );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_131")."\"")
	    unless( $fields{"retry"} =~ /^\d+$/ );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_132")."\"")
	    unless( $fields{"expire"} =~ /^\d+$/ );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_133")."\"")
	    unless( $fields{"ttl"} =~ /^\d+$/ );
	}
    elsif ($type eq 'subdomain' )
	{
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_129")."\"")
	    unless( Cobalt::Netutil::netutil_test_domainname($fields{'ns1'}) );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_129")."\"")
	    if ($fields{"ns2"} && !( Cobalt::Netutil::netutil_test_domainname($fields{'ns2'}) ) );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_135")."\"")
	    if ($fields{'ip1'} && !( Cobalt::Netutil::netutil_test_ipaddr($fields{'ip1'}) ) );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_135")."\"")
	    if ($fields{'ip2'} && !( Cobalt::Netutil::netutil_test_ipaddr($fields{'ip2'}) ) );

	$fields{'value'} = $fields{'ns1'} .':'.
			   $fields{'ip1'} .':'.
			   $fields{'ns2'} .':'.
			   $fields{'ip2'};
	}
    elsif ($type eq 'subnet' )
	{
	my($network_baseip, $network_mask) = split(/\//, $fields{'cdomain'});
	# fields: ip0 (baseip), subnet (slash mask), ns1, ns2 (fqdns)
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui","DNS_04_31")."\"")
	    unless( Cobalt::Netutil::netutil_test_domainname($fields{'ns1'}) );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui","DNS_04_31")."\"")
	    if ($fields{"ns2"} && !( Cobalt::Netutil::netutil_test_domainname($fields{'ns2'}) ) );
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui","DNS_04_30")."\"")
	    unless( netutil_test_network_in_network($fields{'ip0'}, $fields{'subnet'}, $network_baseip, $network_mask ));
	}
    else
	{
	$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_02_61")."\"");
	}

    # If the string isn't empty, return in dismay.
    return 0 if ($user_ERROR ne "" );

    # If we're here, the entry is OK, build up the record

    if ($type eq "host" )
	{ @typelist = ( "a", "ptr" ); }
    else
	{ @typelist = ( $fields{"type"} ); }

    if ( $type eq 'soa' )
    	{
	$rec{ $fields{'rec'} } = 'soa - '. $fields{'cdomain'} .' '.
		   $fields{'ns1'} .':'.
		   $fields{'ns2'} .':'.
		   $fields{'email'} .':'.
		   $fields{'refresh'} .':'.
		   $fields{'retry'} .':'.
		   $fields{'expire'} .':'.
		   $fields{'ttl'} .' -';
    	}
    elsif ( $type eq 'subnet' )
    	{
	$rec{ $fields{'rec'} } = 'subnet - '. $fields{'cdomain'} .' '.
	 	   $fields{'ip0'} .':'.
	 	   $fields{'subnet'} .':'.
		   $fields{'ns1'} .':'.
		   $fields{'ns2'}.' -';
    	}
    else
    	{
        foreach $t ( @typelist )
	    {
	    my $ind = $fields{"rec"};
	    $ind .= "-$t" if ($type eq "host" );

	    $rec{$ind} = $t;
	    for( $_=1; $db_order[$_] ne ""; $_++ )
	        {
	        $f = $db_order[$_];
	        $rec{$ind} .= " " if ($rec{$ind} );
	        $rec{$ind} .= ( ( $fields{$f} ne "" ) ? "$fields{$f}" : "-" );
	        }
	    }
    	}
    &write_records();			# And write it out.
    return 1;
    }

#########################################################################
# Check to see if argument is a valid network name (e.g. 192.168.1).	#
#########################################################################
sub isValidNetworkName
    {
    return 0 if ($_[0] !~ /^[0-9]+(\.[0-9]+)*$/ );
    my( @cpls ) = split(/\./,$_[0]);
    return 0 if ($#cpls < 0 || $#cpls > 3 );
    foreach $_ ( @cpls )
        {
	return 0 if ($_ < 0 || $_ > 255 );
	}
    return 1;
    }

#########################################################################
# Read lines from file into %rec hash.  We clear the array because it	#
# may have been populated by reading from the top of the file.		#
# This might get called again after the file has been modified.		#
#########################################################################
sub read_records
    {
    my $recnum = 0;
    %rec = ();

    if (-e $DNS_records )
	{
	open( INF, $DNS_records ) || die("Cannot open $DNS_records:  $!");
	while( $_ = <INF> )
	    {
	    chomp;
	    $recnum++;
	    $rec{$recnum} = $_;
	    }
	close( INF );
	}
    }

#########################################################################
# Write %rec as lines to a file.					#
#########################################################################
sub write_records
    {

    my( $r, %active_domains, %soa_found, %networks );

    # avoid race
    my $DNS_lock = "$Lockdir/dns-records";
    sleep 1 if (-e $DNS_lock);
    sleep 1 if (-e $DNS_lock);
    sleep 1 if (-e $DNS_lock);
    open( OUT, ">$DNS_lock" ) || die("Cannot write $DNS_lock:  $!");
    foreach $r ( keys %rec )
        {
	next unless ( $rec{$r} );

	my(@record_data) = split(/\s/, $rec{$r});

	if ($record_data[0] eq 'soa')
	    {
	    if($record_data[2] =~ /^[\d\.\/]+$/)
 	        {
	        my($local_ip, $slash_nm) = ($record_data[2] =~ /^([\d\.]+)\/(\d+)/);
		my($local_net) = net_get_network($local_ip, net_convert_netmask($slash_nm)).":$slash_nm";
	        $soa_found{ $local_net } = $rec{$r};
                }
            else
                { # forward domain
		$soa_found{ $record_data[2] } = $rec{$r};
                }
	    next;
	    }
	elsif ($record_data[0] eq 'ptr')
	    {
	    my $ip = $record_data[3];
            my($local_net) = net_get_network($ip, net_convert_netmask($record_data[4])).":$record_data[4]";
	    $active_domains{ $local_net } = $ip.'/'.$record_data[4];
	    }
	elsif (($record_data[0] eq 'subnet') && ($record_data[2] =~ /^([\d\.]+)\/(\d+)/))
	    {
            my($local_net) = net_get_network($1, net_convert_netmask($2)).":$2";
	    $active_domains{ $local_net } = $record_data[2];
	    }
	elsif ($record_data[0] ne 'secondary')
	    {
	    if($record_data[2] =~ /^[\d\.\/]+$/)
 	        {
                my($local_net) =  net_get_network($record_data[2], 
                  net_convert_netmask($record_data[4])).":$record_data[4]";

	        $active_domains{ $local_net } = $record_data[2];
                }
            else
                { # forward domain
		$active_domains{ $record_data[2] } = $record_data[2];
                }
	    }

	print OUT $rec{$r} . "\n";
	}
	
    my( $auth_class );
    foreach $auth_class ( keys %active_domains )
        {
        unless ( $Servers[0] ) # In the loop to save cycles when no SOA records are missing
	    {
            open(NAME,"/proc/sys/kernel/hostname") || die "No hostname: $!";
    	    chomp($Servers[0] = <NAME>);
            close NAME;
	    }

        if ( $soa_found{ $auth_class } )
	    {
	    print OUT $soa_found{ $auth_class } ."\n" if ( $active_domains{ $auth_class } );
	    }
	else
	    {
	    print OUT "soa - ". $active_domains{ $auth_class } ." $Servers[0]::admin\@$Servers[0]:$Refresh:$Retry:$Expire:$Ttl -\n";
	    }
	}

    close( OUT );
    rename($DNS_lock, $DNS_records);

    &read_records();
    }

#########################################################################
# Read an entire file into a string.					#
#########################################################################
sub file_to_string
    {
    my $rmode = $/;
    my $ret = "";
    $/ = "\0777";
    if ( open( INF, $_[0] ) )
	{
	$ret = <INF>;
	close( INF );
	}
    $/ = $rmode;
    return $ret;
    }

#########################################################################
# Print menu that allows user to select zone's format for non 8bit nets	#
#########################################################################
sub zone_format
    {
    my $html = "";
    my $htmlcontents = "";
    my $oldcontents = &file_to_string( $Named_cobdns );
    my %selflags = ();
    my @zonetypes = ();
    opendir( NDD, $Named_dnsdir ) || die("Cannot open $Named_dnsdir:  $!");
    while( $_ = readdir( NDD ) )
	{
	next if ($_ !~ /^$Named_cobprefix.(.*)$/ );
	next if ($1 eq $Named_cobsuffix );
	push( @zonetypes, $1 );
	$_ = &file_to_string("$Named_dnsdir/$_");
	if ($_ eq $oldcontents )
	    {
	    $selflags{$1}="selected";
	    }
	}
    closedir( NDD );
    foreach $_ ( sort @zonetypes )
        {
	$html .= "<option $selflags{$_} value=$_>$_";
	}

    my( $forward0, $forward1, $axs, $r, @rec_data );
    foreach $r ( keys %rec )
	{
	@rec_data = split(/\s+/, $rec{$r});
	next unless ( $rec_data[2] eq '-' ); # Global only
	$forward0 = $rec_data[3] if ( $rec_data[0] eq 'forwarder0' );
	$forward1 = $rec_data[3] if ( $rec_data[0] eq 'forwarder1' );
	$axs = $rec_data[3] if ( $rec_data[0] eq 'axs' );
	}
    $axs =~ s/\:\:/\n/g;

    &templatize( "set_zone_format", "ZONEFMT", $html, 
                 "FORWARD0", $forward0, 
                 "FORWARD1", $forward1,
                 "AXS", $axs );
    return(0);
    }


#########################################################################
# Save global server settings                                           #
#   Zone type, forwarding servers, zone transfer control lists          #
#########################################################################
sub zone_save_status
    {
    # Test data
    $user_ERROR.=eval('"'.Locale::gettext::dgettext("ui", "DNS_04_22").'"')
	if ($fields{'forward0'} && !netutil_test_ipaddr($fields{'forward0'}) );
    $user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_04_23")."\"")
	if ($fields{'forward1'} && !netutil_test_ipaddr($fields{'forward1'}) );

    # The access control list is flexible, formats are either space separated
    # IP addresses or networks in the form "10.10.0.0/16", "192.168.1.200/28"
    if ($fields{'axs'} )
	{
	my( $axs );
	foreach $axs ( split(/[ \n\r]+/,$fields{'axs'}) )
            {
	    my( $axs_ip, $axs_slash, $invalid_network );
	    if ($axs =~ /\/(\d+)$/ )
		{
		$axs_slash = $1;
		}
	
	    $axs_ip = $axs;
	    $axs_ip =~ s/\/$axs_slash//;
	    
	    $invalid_network = 1 unless( netutil_test_ipaddr( $axs_ip ) );

	    if ($axs_slash )
		{
		$invalid_network = 1 if(($axs_slash > 32) || ($axs_slash < 8));
		}

	    if ($invalid_network )
		{
    		$user_ERROR.=eval("\"".Locale::gettext::dgettext("ui", "DNS_04_21")." <B>$axs</B>\"");
		return;
		}
	    }

	$fields{'axs'} =~ s/\s+/\:\:/g;
	}

    my( $forward0, $forward1, $axs, $r, @rec_data, $max_r );
    foreach $r ( keys %rec )
	{
	@rec_data = split(/\s+/, $rec{$r});
	delete( $rec{$r} ) if ($rec_data[2] eq '-' ); # Global only
        $max_r = $r if ($r > $max_r );
	}

    $rec{$max_r+1} = "forwarder0 - - ".$fields{'forward0'}  
	if ($fields{'forward0'} );

    $rec{$max_r+2} = "forwarder1 - - ".$fields{'forward1'}    
	if ($fields{'forward1'} );

    if ($fields{'axs'} )
	{
    	$rec{$max_r+3} = "axs - - ".$fields{'axs'}
	} 
	
    unlink( $Named_cobdns );
    open( OUTF, ">$Named_cobdns") || die("Cannot write $Named_cobdns:  $!");
    print OUTF &file_to_string("$Named_dnsdir/$Named_cobprefix.$fields{zonefmt}");
    close( OUTF );

    &write_records();

    # support automatic vsite configuration.  Nifty
    if ($fields{'autoconfig'}) 
        {
        open(AUTOCONF, "> $DNS_autoconfig"); print AUTOCNF '1'; close(AUTOCONF);
        }
    else 
        {
        unlink($DNS_autoconfig);
        }
    }


#########################################################################
# Convert something to printable under HTML				#
#########################################################################
sub htmlify
    {
    $_ = $_[0];
    s/\&/\&amp/g;
    s/</\&lt/g;
    s/>/\&gt/g;
    return $_;
    }

#########################################################################
# Generate DNS config files from records file				#
#########################################################################
sub update_dns_config
    {
    unless (-e $Named_conf)
        {
        $Named_conf = $Named_disabled; # Edit the disabled configuration instead
        $Offline = 1;
        }

    &generate_domain_contents();
    &clean_up_dir();
    &write_config_files();
    &restart_named();
    }

#########################################################################
# Begin main								#
#########################################################################

sub main {

$q=shift;

my $forbiddenURL="/.cobalt/error/forbidden.html";
my $user = $ENV{REMOTE_USER};
unless (user_authok($user, "admin")) {
    print "location: $forbiddenURL\n\n";
    die "Unauthorized access by $user, not an admin.\n";

}

$user_ERROR = ('');
@Servers = ();    
%domain_contents = ();
%rec = ();           
$Offline = 0; 
$iam = "dns";
my $astring = $q->param( 'func' ); 

print STDERR "String: $astring" if ($DEBUG);

$locale=Cobalt::Locale->getByHTTPHeader( $ENV{ 'HTTP_ACCEPT_LANGUAGE' } );
if ($locale ) { $ENV{ 'LANGUAGE' }=$locale->getLocale();
                $button_style=$locale->getProperty("buttonStyle");
                $header_style=$locale->getProperty("headerStyle");
                $table_style=$locale->getProperty("tableStyle"); }

&read_records();

unless ( ($Features{'bind8'} && (-e $Named_conf)) )
    {
    $Named_conf = $Named_disabled; # Edit the disabled configuration instead
    $Offline = 1;
    }

# %fields = &parseinfo(); 
%fields = $q->Vars; 

my($func) = $fields{"func"};
# my($func) = $astring;

if ($func =~ /show_.*/ )
    {
    &show_get_data($func);
    }
elsif ($func eq "set_zone_format" )
    {
    &zone_format($func);
    }
else
    {
    if ($func eq "update" )
	{
	&update_dns_config();
	print "location: $CallingURL\n\n";
	return(0);
	}
    elsif ($func eq "delete" )
	{
	if ($fields{"rec"} eq 'ALL')
	    {
	    # Delete all records for a domain
	    my($r,@splitRecords);
	    foreach $r (keys %rec) 
                { 
		@splitRecords=split(/\s+/,$rec{$r});
		delete( $rec{$r} ) if 
		  ($splitRecords[2] eq $fields{'cdomain'}); 
		}
            }
    	else
	    {
	    delete( $rec{$fields{"rec"}} );
            }
	&write_records();
	}
    elsif ($fields{"zonefmt"} ne "" )
	{
        &zone_save_status();
 	if ($user_ERROR )
	    {
    	    &zone_format('set_zone_format');
	    return(0);
	    }
        &update_dns_config();
	}
    elsif ($fields{"type"} ne "" )
	{
	if (! &do_modify() )
	    {
	    &show_get_data( $fields{"callback"} );
	    return(0);
	    }
	}
    &show_status();
    }

  undef ($q, $locale);

}

1;
